/** --------------------------------------------------------------------
/** pquotac.c
/** Interactive client for communicating with the pquotad generic
/** quota system. 
/** 
/** Usage: 
/**    pquota [-s server] [-p portnum] [command]
/** 
/** If no command line arguments are given the program will enter an
/** interactive mode.
/** 
/** Commands: See usage function below.
/**
/** Build:
/**   cc -c utils.c
/**   cc -o pquota pquota.c utils -lV3
/**
/** Modification history:
/** Programmer      Date         Description
/** J.Schrock       29-Apr-1994  1.00 Initial design and coding.
/** J.Schrock       16-Sep-1994  1.10 Added comment for STOP command.
/** J.Schrock       21-Mar-1994  2.00 Added new hard quota commands.
/**
/**
/**
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pwd.h>

#include <sys/types.h>
#include <sys/param.h>
#include <signal.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>


#include "pquota.h"

/** --------------------------------------------------------------------
/** main()
/** Main program entry. 
/**
/**
**/

main(argc, argv)
int argc;
char *argv[];
{
   char server_hostname[MAXHOSTNAMELEN];   /* Server hostname. */
   char server_ip[14];                     /* Server IP address. */
   int  server_port=PORTNUM;               /* BSD stream socket */

   char cmd[128];                          /* Command sent to server. */
   char usercmd[128];                      /* Command given by user. */
   char buf[128];                          /* Response from server. */

   struct passwd *pw;                      /* Pointer to password structure */
   int  uid;                               /* User's UID. */
   int  uidroot;                           /* root's UID. */
   int  uidlp;                             /* lp's UID. */
   int  uidlpd;                            /* lpd's UID. */

   int sockfd;                             /* Rendezvoud socket filedes. */
   struct sockaddr_in server;              /* Description of pquotad server */

   struct hostent *host;                   /* Hostname and address */

   int oneshot=0;            /* 1 if command is given on command line. */
   int i;


   /* Copy default server hostname into modifiable memory. */

   strcpy(server_hostname, SERVER);

   /* Parse command line. */

   usercmd[0]=NULL;
   for (i=1; i<argc; i++) {
      if (!strcmp(argv[i], "-s")) {
	 ++i;
	 strcpy(server_hostname, argv[i]);
      }
      else if (!strcmp(argv[i], "-p")) {
	 ++i;
	 server_port = atoi(argv[i]);
      }
      else {
	 sprintf(usercmd, "%s %s", usercmd, argv[i]);
	 oneshot = 1;
      }
   }

   /* Only users root, lp, and lpd can do everything. */
   /* Everyone else is restricted to just a GET. */
   /* This security probably ought to be at the server end. */

   uidroot = 0;
   uidlp   = -1;
   uidlpd  = -1;
   uid     = getuid();

   if ((pw=getpwnam("lp")) != NULL) {
      uidlp  = pw->pw_uid;
   }
   if ((pw=getpwnam("lpd")) != NULL) {
      uidlpd = pw->pw_uid;
   }
   if (uid != uidroot && uid != uidlp && uid != uidlpd) {
      sprintf(usercmd, "GET %d", uid);    
      oneshot = 1;
   }

   /* If we're to do this just once... */

   if (oneshot == 1) {

      if ((host=gethostbyname(server_hostname)) == NULL) {
	 printf("\npquota: Unable to resolve server hostname.\n");
	 exit(1);
      }

      bzero((char*) &server, sizeof(struct sockaddr_in)); 
      bcopy(host->h_addr, (char*) &server.sin_addr, host->h_length);
      strcpy(server_ip, inet_ntoa(server.sin_addr));
      server.sin_family = host->h_addrtype;
      server.sin_port   = server_port;

      if ((sockfd=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	 printf("\npquota: Unable to open stream socket.\n");
	 exit(1);
      }

      if (connect(sockfd, (struct sockaddr *) &server, sizeof(struct sockaddr_in)) < 0) {
	 printf("\npquota: Unable to connect to server\n");
	 exit(1);
      }

      for (i=0; usercmd[i] != ' ' && usercmd[i] != NULL; i++) {
         usercmd[i]=toupper(usercmd[i]);
      }

      sprintf(cmd, "%s %s\n", MAGIC_COOKIE, usercmd); 
      writenet(sockfd, cmd, strlen(cmd));
      getresult(sockfd);
      close(sockfd);
      exit(0);
   }

   /* Otherwise go into an interactive mode. */
   /* Make him use QUIT to get out. */

   sighold(SIGINT);

   while(1) {

      if ((host=gethostbyname(server_hostname)) == NULL) {
	 printf("\npquota: Unable to resolve server hostname.");
      }

      bzero((char*) &server, sizeof(struct sockaddr_in));
      bcopy(host->h_addr, (char*) &server.sin_addr, host->h_length);
      strcpy(server_ip, inet_ntoa(server.sin_addr));
      server.sin_family = host->h_addrtype;
      server.sin_port   = server_port;

      printf("\nServer    : %s", server_hostname);
      printf("\nServer IP : %s  port %d", server_ip, server_port);
      printf("\n");

      /* Prompt for a command. */
      /* Uppercase the first token. */

      printf("\nPQUOTA> ");
      gets(usercmd);
      if (strlen(usercmd) <= 0) {
	 continue;
      }

      for (i=0; usercmd[i] != ' ' && usercmd[i] != NULL; i++) {
	 usercmd[i]=toupper(usercmd[i]);
      }

      if (!strncmp(usercmd, "SERVER", 6)) {
	 sscanf(usercmd, "%s %s", cmd, server_hostname);
      }
      else if (!strncmp(usercmd, "PORTNUM", 7)) {
	 sscanf(usercmd, "%s %d", cmd, &server_port);
      }
      else if (!strncmp(usercmd, "HELP", 4)) {
	 help();
      }
      else if (!strncmp(usercmd, "QUIT", 4)) {
	 break;
      }

      /* Send the command to the server for handling. */

      else {
	 if ((sockfd=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    printf("\npquota: Unable to open stream socket.");
	 }
	 else if (connect(sockfd, (struct sockaddr *) &server, sizeof(struct sockaddr_in)) < 0) { 
	    printf("\npquota: Unable to connect to server.");
	    close(sockfd);
	 }
	 else {
	    sprintf(cmd, "%s %s\n", MAGIC_COOKIE, usercmd); 
	    writenet(sockfd, cmd, strlen(cmd));
	    getresult(sockfd);
	    close(sockfd);
	 }
      }
   }
   exit(0);
}

/** --------------------------------------------------------------------
/** getresult()
/** Reads from the socket until a CTRL-D is seen. Received data is
/** written to stdout. Will assume an error has ocurred if more than
/**
**/
getresult(sockfd)
int sockfd;
{ 
   char buf[126];
   int n;

   while (1) {
      n=readnet(sockfd, buf, 126);
/*      if (buf[0] == '\004' || n <= 0) { */
      if (n == 0) {
	 break;		/* Connection was closed. */
      } 
      printf("%s", buf);
   }
   return(0);
}
   

/** --------------------------------------------------------------------
/** help()
/** Display the command usage.
/**
**/

int help()
{
   printf("\n");
   printf("\nPquota Editor Command Interface");
   printf("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
   printf("\nADD[SOFT] uid amount  - Add amount to volatile quota.");
   printf("\nADDHARD uid amount    - Add amount to permanent quota.");
   printf("\n");
   printf("\nDEDUCT uid amount     - Deduct amount from volatile then from hard quota.");
   printf("\n                        This is the normal way to subtract quota.");
   printf("\nDEDUCTSOFT uid amount - Deduct amount from volatile quota.");
   printf("\nDEDUCTHARD uid amount - Deduct amount from permanent quota.");
   printf("\n");
   printf("\nZERO[SOFT] uid        - Set volatile quota to zero.");
   printf("\nZEROHARD uid          - Set permanent quota to zero.");
   printf("\nZEROALL uid           - Set volatile and permanent quotas to zero.");
   printf("\n");
   printf("\nSET[SOFT] uid amount  - Set volatile quota to amount.");
   printf("\nSETHARD uid           - Set permanent quota to amount.");
   printf("\nSETALL uid            - Set volatile and permanent quotas to amount.");
   printf("\n");
   printf("\nGET uid               - Show hard and soft quotas and their sum.");
   printf("\n");
   printf("\nDELETE uid            - Delete quota entry.");
   printf("\nPROTO uid uid2        - Set hard and soft quota for uid to that of uid2.");
   printf("\nREPORT                - Report on all quota entries.");
   printf("\nWHAT                  - Show database file information.");
   printf("\nSTOP                  - Terminate the pquotad server.");
   printf("\n");
   printf("\nThese commands are serviced locally by this client:");
   printf("\n");
   printf("\nSHOW                  - Display communication settings.");
   printf("\nSERVER hostname       - Change pquotad server hostname.");
   printf("\nPORTNUM socket        - Change pquotad socket number.");
   printf("\nQUIT                  - Quit.");
   printf("\n\n");
}

/** End of file **/	 
