/**********************************************************************
/** pquotad.c v2.00
/** Quota management facility.
/** 
/** This program was designed to keep track of a hard (permanent) and
/** soft (volatile) printing quota associated with user UIDs. The goal
/** was to create a printing quota system. 
/** 
/** The soft/volatile quota is meant to be periodically reassigned or
/** granted by site management, perhaps each semister.
/** 
/** The hard/permanent quota allows a non-resetable quota to be given.
/** For example, this might be quota which is purchased by the user
/** in addition to the granted soft quota.
/** 
/** Pquotad returns both the hard and soft quota values as well as
/** their sum.
/** 
/** 'pquota' and 'pquotad' form a client-server pair. pquotad listens 
/** for and processes requests made by pquota, and it manages the quota
/** database.
/** 
/** The database is made using the Unix standard NDBM.
/**
/** The IPC mechanism is Berkeley Sockets. 
/**
/** This program will not put itself into the background. If that's
/** where you want it envoke the command followed by an ambersand.
/**
/** Usage:  pquotad [-v] quotafile &
/**         
/**    quotafile   Disk file where quotas are kept.
/**    portnum     Socket on which to listen for connection requests.
/**                Default PORTNUM is given in 'pquota.h.ini'.
/**
/** The commands which can be sent to 'pquotad' follow. They have the
/** the names they do because version 1.00 supported just one quota
/** value per UID, and I need to stay compatable with that.
/**
/** Commands:
/**
/**    cookie ADD user amount          Adds amount to volatile quota.
/**    cookie ADDSOFT user amount      Adds amount to volatile quota.
/**    cookie ADDHARD user amount      Adds amount to permanent quota.
/**    cookie ADDALL user amount       Adds amount to volatile & permanent quota.
/**
/**    cookie DEDUCT user amount       Subtracts amount first from volatile quota 
/**                                    then from permanent quota. This is the normal
/**                                    way to deduct pages from a user's quota.
/**    cookie DEDUCTSOFT user amount   Subtracts amount from volatile quota.
/**    cookie DEDUCTPERM user amount   Subtracts amount from permanent quota.
/**
/**    cookie SET user amount          Sets volatile quota to amount.
/**    cookie SETSOFT user amount      Sets volatile quota to amount.
/**    cookie SETPERM user amount      Sets permanent quota to amount.
/**    cookie SETALL user amount       Sets volatile & permanent quota to amount.
/**
/**    cookie PROTO user user2         Sets quotas to be the same as uid2's.
/**    cookie GET user                 Returns current quota balances. 
/**
/**    cookie ZERO user                Sets volatile quota to zero.
/**    cookie ZEROSOFT user            Sets volatile quota to zero.
/**    cookie ZEROPERM user            Sets permanent quota to zero.
/**    cookie ZEROALL user             Sets volatile & permanent quota to zero.
/**
/**    cookie DELETE user              Deletes quota entrie for uid.
/**    cookie REPORT                   Reports on all defined quota entries.
/**    cookie STOP                     Gracefully stops pquotd.
/**
/**    Where:
/**   
/**    amount                          An integer value.
/**    user & user2                    User UID number or valid usernames.
/**    cookie                          A magic cookie string.
/**
/** Completion codes:
/**    Each call to this server generates a response. Each response is 
/**    prefixed with one of these condition codes. A text message follows 
/**    each condition code on the same record. Each complete response is 
/**    followed by a ^D (0x004) character. Clients calling this function 
/**    must continue to read the responses until the ^D is seen.
/**
/**    000  Quota for UID %d (%s) is %d
/**         Normal reponse to all operations except DELETE and REPORT.
/**
/**    000  Quota for UID %d (%s) was deleted
/**         Normal response to a DELETE operation.
/**
/**    000  pquotad terminating on command STOP.
/**         Normal response to a STOP operation.
/**
/**    000  Database was created %s
/**    000  Database was last modified %s
/**         Normal response to a WHAT command.
/**
/**    8XX  Error deleting quota for UID %d (%s)
/**         A call to dbm_delete() failed for the reason indicated.
/**
/**    8XX  Error updating quota for UID %d (%s)
/**         A call to dbm_store() failed for the reason indicated.
/**
/**    901  Unable to open quota database
/**         A call to dbm_open() failed. The filename prefix given on the command 
/**         line could not be opened as a NDBM database for read/write access.
/**
/**    902  Quota for prototype UID %d (%s) is undefined
/**         The UID given as a prototype did not have a quota entry.
/**
/**    903  Quota for UID %d (%s) is undefined
/**         The DELETE operation failed because the UID does not have a quota.
/**  
/**    904  Quota for UID %d (%s) is undefined
/**         The GET operation failed because the UID does not have a quota.
/**
/**    905  Username %s is unknown.
/**         A specified username is not valid.
/**
/**    906  Username %s is unknown.
/**         A username given as a prototype is not valid.
/**
/**    907  Internal error validating quota for UID %d (%s)
/**
/**    908  Internal error retrieving creation date.\n");
/**         The date the database was created could not be determined.
/**
/**    909  Internal error retrieving modification date.\n");
/**         The date the database was created could not be determined.
/**
/**    999  Invalid command
/**         The command given is invalid.
/**
/**
/** Authentication:
/**    For the moment authentication is based on the client prefixing
/**    each server request with the MAGIC_COOKIE.  Something better
/**    needs to be done in the future.
/**
/** Build:
/**    cc -c utils.c
/**    cc -o pquotad pquotad.c -lndbm -lV3
/**
/** Reference:
/**    "UNIX Network Programming", W. Richard Stevens, Prentice Hall, 1990.
/**
/**
/** Modification history:
/** Programmer      Date         Description
/** J.Schrock       29-Apr-1994  1.00 Initial design and coding.
/** J.Schrock       16-Sep-1994  1.10 Added STOP command.
/** J.Schrock       13-Mar-1996  2.00 Added support for permanent quotas.
/**                                   Added syslog support.
/**
/**
**/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ndbm.h>
#include <syslog.h>

#include <pwd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <signal.h>

#include <time.h>

#include <errno.h>

#include "pquota.h"

/** --------------------------------------------------------------------
/** Global variables.
/**
**/

int  sockfd;                 /* Rendezvous socket. */
int  newsockfd;              /* Communications socket. */
int  logging=LOGGING;	     /* Logging level. */
int  portnum=PORTNUM;        /* BSD socket number. */
int  istatus;                /* Event completion status. */
int  event=10000;            /* Event number. */

/** --------------------------------------------------------------------
/** Main Program.
/** Set ourselves up as a daemon and start a service connection.
/** 
**/

main(argc, argv)
int argc;
char *argv[];
{
   struct sockaddr_in cli_addr; /* Socket defining the client. */
   struct sockaddr_in serv_addr;/* Socket defining the server. */

   int clilen;                        

   struct hostent *peer;       /* Name of the peer process's host. */

   char result[80];            /* String sent out on the net. */
   char string[128];           /* Miscellaneous string. */

   char inbuf[128];            /* Stream buffer. */
   int  MAXBUF=128;       

   char action[10];            /* Action to take. */

   char user[80];              /* Username or UID of quota to modify. */
   int  uid;                   /* UID of quota to modify. */

   char argument[80];          /* Argument to action request. */
   int  amount;                /* Amount of modification or second UID. */

   char quotafile[1024];       /* Name of quota database file. */
   char magic[80];             /* Magic cookie. */

   struct passwd *pw;          /* Pointer to password structure. */

   DBM *db;	               /* File pointer to DBM database. */
#ifndef __hpux
   DBM *dbm_open();
#endif

   extern void sighandler();   /* Interrupt service routine. */

   int i;		       /* Miscellaneous. */

   /* Gotta be root! */

   if (getuid() != 0) {
      printf("\npquotad: Must be root!\n");
      exit(0);
   }

   /* Parse command line. */

   if (argc == 3) {
      if (strcmp(argv[1],"-v")) {
	 logging=1;
      }
      else if (strcmp(argv[1],"+v")) {
	 logging=0;
      }
      else {
	 printf("\nusage: pquotad [+/-v] database_filename\n");	 
      }
   }
   else if (argc == 2) {
      strcpy(quotafile, argv[1]);
   }	
   else {
      printf("\nusage: pquotad [+/-v] database_filename\n");
      exit(0);
   }

   /* Try to open a line to the syslogger. */

   if (logging) {
      openlog("pquotad: ", LOG_PID, LOG_LOCAL6);
   }

   /* Don't allow a wayward client to hose us. */
   /* If a client breaks ties with us too soon it will cause */
   /* a "broken pipe" (SIGPIPE) signal to be generated. */

   sighold(SIGPIPE);

   /* Open a new database if one doesn't already exist. */
   /* If it's new, insert a creation date time stamp. */

   if ((db=dbm_open(quotafile, O_RDONLY, 700)) == NULL) {
      if ((db=dbm_open(quotafile, O_RDWR | O_CREAT, 700)) == NULL) {
         sprintf(string, "Failed to create and/or open database: %s", strerror(errno));
	 logger(string, 0);
	 return(1);
      }
      timestamp(db, CREATE_TIME_KEY);
   }
   dbm_close(db);
   
   sprintf(string, "Starting up using database file %s.", quotafile);
   logger(string, 0);

   /* Open a socket. */

   if ((sockfd=socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      sprintf(string, "Socket open failed: %s", strerror(errno));
      logger(string, 0);
      exit(2);
   }

   /* Register this socket with the system. */

   for (i=0; i<10; ++i) {
      bzero((char *) &serv_addr, sizeof(struct sockaddr_in)); 
      serv_addr.sin_family      = AF_INET;
      serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
      serv_addr.sin_port        = htons(portnum);

      if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(struct sockaddr_in)) < 0) {
	 sprintf(string, "Socket bind failed: %s", strerror(errno));
	 logger(string, 0);
	 sprintf(string, "Will retry socket bind in 15 seconds.");
	 logger(string, 0);
	 sleep(15);
	 sprintf(string, "Retrying socket bind.");
	 logger(string, 0);
	 continue;
      }
      break;
   }
   if (i == 10) {
      logger("Socket bind failed 10 times. Giving up.", 0);
      exit(2);
   }

   /* Setup the listener for requests. */

   if (listen(sockfd, 5) < 0) {
      sprintf(string, "Listen on socket failed: %s", strerror(errno));
      logger(string, 0);
      exit(2);
   }      

   /* Trap all death signals, when allowed to, from here on. */

   signal(SIGINT, sighandler);
   signal(SIGQUIT, sighandler);
   signal(SIGHUP, sighandler);

   /* Now wait for requests to come in. */

   sprintf(string, "Now accepting requests on socket %d.\n", portnum);
   logger(string, 0);

   clilen = sizeof(cli_addr);
   while (1) {
      if ((newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen)) < 0) {
	 sprintf(string, "Server accept failed: %s", strerror(errno));
	 logger(string, 0);
	 exit(2);
      }

      /* This is a real event now. */

      event = ++event;

      /* Reset variables. */

      strcpy(action,"");
      strcpy(argument,"");
      uid=-1;
      amount=-1;

      /* Don't stop here! */

      sighold(SIGQUIT);
      sighold(SIGINT);
      sighold(SIGHUP);

      /* Read the message and act on it iteratively. */
      /* Just continue if nothing no command was given. */

      strcpy(magic,   "");
      strcpy(action,  "");
      strcpy(user,    "");
      strcpy(argument,"");

      readnet(newsockfd, inbuf, sizeof(inbuf));
      sscanf(inbuf, "%s %s %s %s", magic, action, user, argument);

      if (strlen(action) == 0) {
	 continue;
      }	

      /* If the correct magic cookie is not given, do nothing. */
      
      if (strcmp(magic, MAGIC_COOKIE)) {
	 logger("Bad magic cookie given.", 0);
	 goto finish;
      }

      /* Resolve 'user' into a UID. */
      /* If a username was given it must be a valid one. */

      if (strlen(user) > 0) {
	 if (isdigit(user[0])) {
	    uid = atoi(user);
	 }
	 else if ((pw=getpwnam(user)) != NULL) {
	    uid = pw->pw_uid;
	 }
	 else {
	    sprintf(result, "905 Username %s is unknown\n", user);
	    writenet(newsockfd, result, strlen(result));
	    goto finish;
	 }
      }

      /* If 'argument' is an alpha string then it must be a username. */
      /* Get its corrosponding UID, too. */

      if (strlen(argument) > 0) {
	 if (isdigit(argument[0])) {
	    amount = atoi(argument);
	 }
	 else if ((pw=getpwnam(argument)) != NULL) {
	    amount = pw->pw_uid;
	 }
	 else {
	    sprintf(result, "906 Username %s is unknown\n", argument);
	    writenet(newsockfd, result, strlen(result));
	    goto finish;
	 }
      }

      /* Open the NDBM database for modification. */




/* 
 *    cc choked on this since db is an int pointer - 
 *    which is no good for a logical comarison with "<"
 *
 *     if ((db=dbm_open(quotafile, O_RDWR | O_CREAT, 700)) < 0) {
 */
        db=dbm_open(quotafile, O_RDWR | O_CREAT, 700);
        if ( (int)db < 0) 
          {


	strcpy(result, "901 Unable to open quota database.\n");
	writenet(newsockfd, result, strlen(result));
	logger("901 Unable to open quota database.", 0);
	goto finish;
      }

      modify_quota(newsockfd, db, quotafile, action, uid, amount);
      dbm_close(db);

      /* Clean up. */
      /* A Control-D signifies the end of our transmission. */

      finish:
      /* writenet(newsockfd, "\004", 1); */
	 close(newsockfd);
	 sigrelse(SIGQUIT);
	 sigrelse(SIGINT);
	 sigrelse(SIGHUP);

   } /* while(1) */
}

/** --------------------------------------------------------------------
/** modify_quota()
/** Function that actually opens and modifies the quota file.
/** 
**/

modify_quota(fd, db, quotafile, action, uid1, amount)
int  fd;
DBM  *db;
char *quotafile;
char *action;
int  uid1;
int  amount;
{
   int arg;                      /* Command argument. */
   int cmd;                      /* Quota command. */
   int flag;                     /* NDBM update or insert flag. */
   int err;                      /* Return code for NDBM accesses */
   int i;

   int create_time_key=CREATE_TIME_KEY;
   int modify_time_key=MODIFY_TIME_KEY;

   char result[128];             /* String returned to caller. */
   char string[128];		 /* Miscellaneous string. */

   struct passwd *pw;		 /* Pointer to password entry. */
   char username[60];            /* Username associated with UID1. */

   static datum key;             /* Database key. */
   static datum quotaent1;       /* Database fetched item for UID1. */
   static datum quotaent2;       /* Database fetched item for UID2. */

   struct quotaitem {		 /* Breaks datum into hard and soft quotas. */
      int soft_quota;
      int hard_quota;
   } *quotaptr1, *quotaptr2, quotadata;

   int newhardamount; 		 /* New soft quota value to be set. */
   int newsoftamount; 		 /* New hard quota value to be set. */

   /* No errors seen yet. */

   istatus=0;

   /* Determine the username associated with UID1. */

   if ((pw=getpwuid(uid1)) == NULL) {
      strcpy(username, "unknown");
   }
   else {
      strcpy(username, pw->pw_name);
   }

   /* Determine the amount a quota needs to be adjusted. */
   /* ADD operation: Add amount to new or existing soft quota. */
   
   if (!strcmp(action, "ADD") || !strcmp(action, "ADDSOFT")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      /* Quota does not exist yet. */

      if (quotaptr1 == NULL) {	
	 newsoftamount = amount;
	 newhardamount = 0;
      }

      /* Quota does exist. Add to the soft quota. */

      else  {
	 newsoftamount = quotaptr1->soft_quota + amount;
	 newhardamount = quotaptr1->hard_quota;
      }
   }

   /* ADDHARD operation: Add amount hard quota. */
   
   else if (!strcmp(action, "ADDHARD")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      /* Quota does not exist yet. */

      if (quotaptr1 == NULL) {	
	 newsoftamount = 0;
	 newhardamount = amount;
      }
      else  {
	 newsoftamount = quotaptr1->soft_quota;
	 newhardamount = quotaptr1->hard_quota + amount;
      }
   }

   /* DEDUCT operation: Subtract first from soft quota, then hard. */

   else if (!strcmp(action, "DEDUCT")) {
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = -amount;
	 newhardamount = 0;
      }
      else {
	 newsoftamount = quotaptr1->soft_quota - amount;
         if (newsoftamount < 0) {
	    newhardamount = quotaptr1->hard_quota + newsoftamount;
	    newsoftamount = 0;
         }
         else {
	    newhardamount = quotaptr1->hard_quota;
	 }
      }
   }

   /* DEDUCTHARD operation: Subtract from hard quota only. */

   else if (!strcmp(action, "DEDUCTHARD")) {
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = 0;
	 newhardamount = -amount;
      }
      else {
	 newsoftamount = quotaptr1->soft_quota;
	 newhardamount = quotaptr1->hard_quota - amount;
      }
   }

   /* DEDUCTSOFT operation: Subtract from soft quota only. */

   else if (!strcmp(action, "DEDUCTSOFT")) {
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = -amount;
	 newhardamount = 0;
      }
      else {
	 newsoftamount = quotaptr1->soft_quota - amount;
	 newhardamount = quotaptr1->hard_quota;
      }
   }

   /* SET[SOFT] operation: Set soft quota amount. */

   else if (!strcmp(action, "SET") || !strcmp(action, "SETSOFT")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = amount;
	 newhardamount = 0;
      }
      else {
	 newsoftamount = amount;
	 newhardamount = quotaptr1->hard_quota;
      }
   }

   /* SETHARD operation: Set permanent quota amount. */

   else if (!strcmp(action, "SETHARD")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = 0;
	 newhardamount = amount;
      }
      else {
	 newsoftamount = quotaptr1->soft_quota;
	 newhardamount = amount;
      }
   }

   /* SETALL operation: Set hard and soft quotas to amount. */

   else if (!strcmp(action, "SETALL")) {	
      newsoftamount = amount;
      newhardamount = amount;
   }

   /* ZERO[SOFT] operation: Set the soft quota to zero. */

   else if (!strcmp(action, "ZERO") || !strcmp(action, "ZEROSOFT")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = 0;
	 newhardamount = 0;
      }
      else {
	 newsoftamount = 0;
	 newhardamount = quotaptr1->hard_quota;
      }
   }

   /* ZEROHARD operation: Set the hard quota to zero. */

   else if (!strcmp(action, "ZEROHARD")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      quotaptr1 = quotaent1.dptr;

      if (quotaptr1 == NULL) {
	 newsoftamount = 0;
	 newhardamount = 0;
      }
      else {
	 newsoftamount = quotaptr1->soft_quota;
	 newhardamount = 0;
      }
   }

   /* ZEROALL operation: Set the hard and soft quotas to zero. */

   else if (!strcmp(action, "ZEROALL")) {	
      newsoftamount = 0;
      newhardamount = 0;
   }

   /* PROTO operation: Copy hard and soft quotas from another UID. */

   else if (!strcmp(action, "PROTO")) {
      key.dptr  = (char*)&amount;
      key.dsize = sizeof(int);
      quotaent2 = dbm_fetch(db, key);
      quotaptr2 = quotaent2.dptr;

      if (quotaptr2 == NULL) {
	 sprintf(result, "902 Quota for prototype UID %d (%s) is undefined.\n",
		 uid1, username);
	 writenet(fd, result, strlen(result));
	 istatus=902;
	 goto byebye;
      }
      newsoftamount = quotaptr2->soft_quota;
      newhardamount = quotaptr2->hard_quota;
   }


   /* DELETE operation: Remove quota entry for UID. */

   else if (!strcmp(action, "DELETE")) {
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);

      if (quotaent1.dptr != NULL) {
	 dbm_delete(db, key);
	 if ((err=dbm_error(db)) != 0) {
	    dbm_clearerr(db);
	    sprintf(result, "8%2.2d Error deleting quota entry for UID %d (%s)\n.", 
		    err, uid1, username);
	    istatus=800 + err;
	 }
	 else {
	    sprintf(result, "000 Quota for UID %d (%s) was deleted.\n",
		   uid1, username);
	 }
      }
      else {
	 sprintf(result, "903 Quota for UID %d (%s) is undefined.\n",
		 uid1, username);
	 istatus=903;	
      }
      writenet(fd, result, strlen(result));
      goto byebye;
   }

   /* GET operation: Return hard and soft quotas and their sum. */

   else if (!strcmp(action, "GET")) {	
      key.dptr  = (char*)&uid1;
      key.dsize = sizeof(int);
      quotaent1 = dbm_fetch(db, key);

      if (quotaent1.dptr == NULL) {
	 sprintf(result, "904 Quota for UID %d (%s) is undefined.\n",
		 uid1, username);
	 istatus=904;
      }
      else {
	 quotaptr1 = quotaent1.dptr;
	 amount    = quotaptr1->soft_quota + quotaptr1->hard_quota;
	 sprintf(result, "000 Quota for UID %d (%s) is %d (soft=%d, perm=%d)\n", 
		 uid1, username, amount, quotaptr1->soft_quota, quotaptr1->hard_quota);
      }
      writenet(fd, result, strlen(result));
      goto byebye;
   }

   /* REPORT operation. */

   else if (!strcmp(action, "REPORT")) {
      writenet(fd, "000  User-ID     Username     Quota        Soft        Hard    \n", 64);
      writenet(fd, "000 ----------  ----------  ----------  ----------  ---------- \n", 64);
                                              
      for (key=dbm_firstkey(db); key.dptr != NULL; key=dbm_nextkey(db)) {
	 quotaent1 = dbm_fetch(db, key);
	 quotaptr1 = quotaent1.dptr;

	 if ((pw=getpwuid(*(int*)key.dptr)) == NULL) {
	    strcpy(username, "unknown");
	 }
	 else {
	    strcpy(username, pw->pw_name);
	 }

	 amount=quotaptr1->soft_quota + quotaptr1->hard_quota;

	 sprintf(result, "000 %10d  %10s  %10d  %10d  %10d \n", 
		 *(int*)key.dptr, username, amount, quotaptr1->soft_quota, quotaptr1->hard_quota);
	 writenet(fd, result, strlen(result));
      }
      goto byebye;
   }

   /* STOP command. */

   else if (!strcmp(action, "STOP")) {
      writenet(fd, "000 pquotad terminating on command STOP.\n", 41);
      close(sockfd);
      close(newsockfd);
      logger("Terminated on command STOP.", 0);
      exit(1);
   }

   /* WHAT operation. */
  
   else if (!strcmp(action, "WHAT")) {
      sprintf(result, "000 Database file is %s\n", quotafile);
      writenet(fd, result, strlen(result));

      key.dptr = (char*)&create_time_key;
      key.dsize= sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      if (quotaent1.dptr == NULL) {
	 sprintf(result, "908 Internal error retrieving creation date.\n");
	 writenet(fd, result, strlen(result));
	 istatus=908;
      }
      else {
	 sprintf(result, "000 Database was created %s\n", quotaent1.dptr);
  	 writenet(fd, result, strlen(result));   
      }

      key.dptr = (char*)&modify_time_key;
      key.dsize= sizeof(int);
      quotaent1 = dbm_fetch(db, key);
      if (quotaent1.dptr == NULL) {
	 sprintf(result, "809 Internal error retrieving modification date.\n");
	 writenet(fd, result, strlen(result));
	 istatus=809;
      }
      else {
	 sprintf(result, "000 Database was last modified %s\n", quotaent1.dptr);
	 writenet(fd, result, strlen(result));   
      }
      goto byebye;
   }

   /* Bad command. */

   else {
      sprintf(result, "999 Invalid command\n");
      writenet(fd, result, strlen(result));
      istatus=999;
      goto byebye;
   }

   /* We made it this far, so we are going to modify the database. */
   /* Time stamp the event. */

   timestamp(db, MODIFY_TIME_KEY);

   /* Find out if there is already a quota for this UID. */
   /* If there is, update it; otherwise, insert a new one. */

   key.dptr  = (char*)&uid1;
   key.dsize = sizeof(int);
   quotaent1 = dbm_fetch(db, key);
   flag = (quotaent1.dptr == NULL)? DBM_INSERT : DBM_REPLACE ;

   /* Modify the quota entry. */

   key.dptr = (char*)&uid1;
   key.dsize= sizeof(int);

   quotadata.soft_quota = newsoftamount;  
   quotadata.hard_quota = newhardamount;  

   quotaent1.dsize = 2 * sizeof(int);
   quotaent1.dptr  = &quotadata;

   dbm_store(db, key, quotaent1, flag);
   if ((err=dbm_error(db)) != 0) {
      dbm_clearerr(db);
      sprintf(result, "8%2.2d Error updating quota for UID %d (%s)\n",
	      err, uid1, username);
      writenet(fd, result, strlen(result));
      istatus=800 + err;
      goto byebye;
   }

   /* Debugging. Retrieve the just set quota. */

   key.dptr = (char*)&uid1;
   key.dsize= sizeof(int);

   quotaent1 = dbm_fetch(db, key);
   quotaptr1 = quotaent1.dptr;

   if (quotaptr1 == NULL) {
      sprintf(result, "907 Internal error validating quota for UID %d (%s)\n",
	      uid1, username);
      writenet(fd, result, strlen(result));
      istatus=907;
      goto byebye;
   }

   amount = quotaptr1->soft_quota + quotaptr1->hard_quota;
   sprintf(result, "000 Quota for UID %d (%s) is %d (soft=%d, perm=%d)\n", 
	   uid1, username, amount, quotaptr1->soft_quota, quotaptr1->hard_quota);
   writenet(fd, result, strlen(result));

   /* This is the way out of this function. */

 byebye:

   sprintf(string, "Event=%d Cmd=%-8.8s User=%-8.8s Status=%d", 
	   event, action, username, istatus);
   logger(string, 1);
   return(0);
}

/** --------------------------------------------------------------------
/** --- timestamp(db, tkey);
/** --- Puts a time entry into the database using 'tkey' as the
/** --- record key.
/** ---
/** ---
**/
int timestamp(db, tkey)
DBM *db;
int  tkey;
{
   int    err;
   time_t tloc;                  /* Timer ticks */
   char timestr[30];             /* Date and time string. */
   char string[128];		 /* Miscellaneous string. */

   static datum key;             /* Database key. */
   static datum quotaent;        /* Database item. */

   time(&tloc);   
   strftime(timestr, 30, "%d-%b-%Y %H:%M:%S", localtime(&tloc));

   key.dptr = (char*)&tkey;
   key.dsize= sizeof(int);
   quotaent.dptr = timestr;
   quotaent.dsize= strlen(timestr);

   dbm_store(db, key, quotaent, DBM_REPLACE);
   if ((err=dbm_error(db)) != 0) {
     dbm_clearerr(db);
     sprintf(string, "Error 8%2.2d, failed to timestamp database.\n", err);
     logger(string, 0);
     return(1);
   }
   return(0);
}

/** --------------------------------------------------------------------
/** logger(string, flag)
/** Writes a string to syslogd and optionally if flag=1 to stdout.
/**
**/
logger(string, flag)
char *string;
int  flag;
{
   if (flag <= logging) {
      printf("\npquotad: %s", string);
      syslog(LOG_INFO, "%s", string);
   }
}

/** --------------------------------------------------------------------
/** sighandler()
/** Called to trap various death notifications.
/** Closes the TCP sockets.
/**
**/

void sighandler(arg)
int arg;
{
   close(sockfd);
   close(newsockfd);
   logger("Terminating on trapped signal.", 0);
   exit(1);
}

/** End of file **/	 
